DEFINT A-Z

                FUNCTION Xinput$ (Row%, Col%, FieldLen%, Default$, AllowCharsMask$, EndingKeys$, ExitKey$, RealStrLen%) STATIC

'Ŀ
'         XINPUT 2.1:  A Powerful Replacement for standard QBasic INPUT       
'                 
'                              First written                                  
'                                   by                                        
'          Frederick Volking, QBNews, Volume  1, Number  3  May 22, 1990      
'                                                                             
'                    ͻ                      
'                             Paolo Bagnaresi                               
'                           Via J.F. Kennedy 17                             
'                      20097 San Donato Milanese (MI)                       
'                                Italy                                      
'                          Tel. ++39-2-514.202                              
'                                                     
'                    e-mail: paul.bagnaresi@gmail.com 
'                    ͼ                      
'                                                                             
'          Almost entirely rewrote it and deeply enhanced it on June 1995     
'                                                                             
' Vers. 2.0 : Some bad bugs corrected on March 1996 (cursor position and      
'             returned string length)                                         
' Vers. 2.1 : May 2000. FieldLen% bug fixed.                                  
'Ĵ
'SYNTAX:Xinput$(Row%,Col%,FLen%,Default$,Allow$,EndKeys$,ExitKey$,RealStrLen%)
'WHERE: Row% & Col% - where input to occur                                    
'         FLen% - Field Length - maximum allowable                            
'         Default$ - Field's beginning default value                          
'         Allow$ - Character Set allowable for input                          
'         EndKeys$ - Keystrokes acceptable for finishing.                     
'            SUCH: [Enter] is CHR$(13)                                        
'                  [Esc] is CHR$(27)                                          
'                  [F1] is CHR$(255)+CHR$(59)                                 
'                  [Ctrl+PgUp] is CHR$(255)+CHR$(132)                         
'                  ... etc                                                    
'         ExitKey$ - The keystroke (from above EndKeys$)                      
'                    which was struck at function end.                        
'         RealStrLen% - The Real String Length, not counting unused space     
'                       at the string end. However, Spaces added manually     
'                       do count!                                             
'  SIMPLEST USE:  A$ = Xinput$(0,0,0,"","","",Any$,RealStrLen%)               
'         Will default to cursor's current row/col, length is                 
'         autoset to right margin, allowable chars are ALL,                   
'         only ending key is [Enter].                                         
'Ĵ
'                                                                             
'   Instructions and enhancements discussion after the end of program         
'                           ***  see below ***                                
'                                                                             
'
'  Discussion 
'
'                 XINPUT program eXtends the standard QBasic INPUT.
'
'     Unlike QB INPUT, XINPUT allows editing a previously created string.
'     This is the strongest feature of Xinput and the very reason why this
'     task was undertaken. I just could not bear any longer the tedious
'     retyping of any string change that was needed in some program
'     I developed through the years.
'     The original routine written by Frederick Volking on QBNews, named
'     ENPUT,  did not alllow a string editing, which is why I felt compelled
'     to rewrite the Volking routine.
'
'     Other valuable features are:
'
'- Exit keys (Xinput terminators) are user selectable, and detectable by caller.
'- All the user allowed keys are also pre-selectable. Caller can create a mask
'  containing the all the keys that the callee will be allowed to use.
'  A low sound will signal hitting an unselected key.
'
'- Insert key remains active until you deselect it by pressing it again.
'- Backspace pressed 8 times at 1st string position reverts you to start string.
'- Screen start position is user selectable.
'- Cursor Up key and Cursor Down key will move you through the rows.
'- Page Up will moves you to first row. Page Down to the last row.
'- CTRL+Home erases all the field from cursor position up to 1st string char.
'- All the standard QB Input features are emulated, including CTRL+End,
'  CTRL+R, CTRL+E, Tab and CTRL+I...
'- Escape key or CTRL+U, if aren't redefined as an exit key, will entirely
'  blank the string field.
'- Source code in plain QBasic is here, waiting for your enhancements.
'  BTW, I would gladly receive your enhancements, if you can send them to my
'  address:                   Paolo Bagnaresi
'                           Via J.F. Kennedy 17
'                      20097 San Donato Milanese (MI)
'                                Italy
'                          Tel. ++39-2-514.202
'                   e-mail: paul.bagnaresi@gmail.com 

                SHARED ScreenWidth%, EmptySpaceChar$

' 1 character Key vakues
                CtrlB$ = CHR$(2)         ' Ctrl+ B = Cursor to the word on the left
                CtrlE$ = CHR$(5)         ' Ctrl+ E = Cancel all chars from Cursor Posit. to end
                CtrlF$ = CHR$(6)         ' Ctrl+ F = Cursor to the word on the right
                BackS$ = CHR$(8)         ' Backspace
                Tabul$ = CHR$(9)         ' Tabulator
                CtrlI$ = CHR$(9)         ' Ctrl+ I = Tabulator
                CtrlK$ = CHR$(11)        ' Ctrl+ K = Cursor to start of field (= Home)
                Enter$ = CHR$(13)        ' Enter
                CtrlN$ = CHR$(14)        ' Ctrl+ N = Cursor to end of field (= End)
                CtrlR$ = CHR$(18)        ' Ctrl+ R = Insert mode toggle
                CtrlU$ = CHR$(21)        ' Ctrl+ U = Clear all input field
                Esc$ = CHR$(27)          ' Escape
                CtrlRevSlash$ = CHR$(28) ' Ctrl+ \ = Cursor Right
                CtrlClSquare$ = CHR$(29) ' Ctrl+ ] = Cursor Left
                                
                CTRLC$ = CHR$(3)	      ' CTRL+ C
                CTRLV$ = CHR$(22)        ' CTRL+ V

' 2 character Key vakues
                CursLeft$ = CHR$(0) + CHR$(75)
                CursRight$ = CHR$(0) + CHR$(77)
                CursUp$ = CHR$(0) + CHR$(72)
                CursDown$ = CHR$(0) + CHR$(80)

                Insert$ = CHR$(0) + CHR$(82)
                Delete$ = CHR$(0) + CHR$(83)
                Home$ = CHR$(0) + CHR$(71)
                End$ = CHR$(0) + CHR$(79)
                PageUp$ = CHR$(0) + CHR$(73)
                PageDown$ = CHR$(0) + CHR$(81)

                CtrlHome$ = CHR$(0) + CHR$(119)
                CtrlEnd$ = CHR$(0) + CHR$(117)
                CtrlPageUp$ = CHR$(0) + CHR$(132)
                CtrlPageDown$ = CHR$(0) + CHR$(118)
                CtrlCursLeft$ = CHR$(0) + CHR$(115)
                CtrlCursRight$ = CHR$(0) + CHR$(116)


' Other Key vakues, currently unused
                F1$ = CHR$(59)           'F1
                F2$ = CHR$(60)           'F2
                F3$ = CHR$(61)           'F3
                F4$ = CHR$(62)           'F4
                F5$ = CHR$(63)           'F5
                F6$ = CHR$(64)           'F6
                F7$ = CHR$(65)           'F7
                F8$ = CHR$(66)           'F8
                F9$ = CHR$(67)           'F9
                F10$ = CHR$(68)          'F10
                F10$ = CHR$(133)         'F11
                F10$ = CHR$(134)         'F12

' Verify if SHARED globals are set
                IF ScreenWidth% = 0 THEN ScreenWidth% = 80
                IF LastRow% = 0 THEN LastRow% = 24
                TotalScreenLength% = LastRow% * ScreenWidth%
                IF EmptySpaceChar$ = "" THEN EmptySpaceChar$ = CHR$(32)

' If not specified then supply defaults to incoming vars
                IF AllowCharsMask$ = "" THEN AllowCharsMask$ = CHR$(34) + " !#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\]^_`abcdefghijklmnopqrstuvwxyz{|}~"
                IF EndingKeys$ = "" THEN EndingKeys$ = CHR$(13)
                IF Row% = 0 THEN Row% = CSRLIN
                IF Col% = 0 THEN Col% = POS(0)
                ReturnVar$ = Default$
                IF FieldLen% = 0 THEN FieldLen% = LEN(ReturnVar$)
                IF FieldLen% <= LEN(ReturnVar$) THEN GOTO FieldLenOk
                ReturnVar$ = ReturnVar$ + STRING$(FieldLen% - LEN(ReturnVar$), EmptySpaceChar$)
                FieldLen% = LEN(ReturnVar$)
                Default$ = ReturnVar$
                
FieldLenOk:
' Define internal defaults

                CursPos% = 1
                ReturnVar$ = Default$
                Separator$ = " !#$%&'()*+,-./:;<=>?@[\]^_`{|}~" + CHR$(34) ' Valid separators for CTRL+LEFT or CTRL+RIGHT
                IF TabLen% = 0 THEN TabLen% = 8 ' Tabulator lenght
                NotDone% = 1
                standard% = 6 'Cursor Shape if in Standard Mode
                Insert% = 4   'Cursor Shape if in Insert Mode
                CurShape% = standard%

'  MAIN LOOP STARTS HERE 
                Riga% = Row%: Lonna% = Col%
          DO
                LOCATE Riga%, Lonna%, 0
                IF Lonna% + LEN(ReturnVar$) > ScreenWidth% THEN
                Part1$ = LEFT$(ReturnVar$, ScreenWidth% - Lonna% + 1)
                Part2$ = RIGHT$(ReturnVar$, LEN(ReturnVar$) - (ScreenWidth% - Lonna% + 1))
                ELSE
                Part1$ = ReturnVar$: Part2$ = ""
                END IF
                PRINT Part1$; Part2$;

                RowCurs% = INT((Lonna% + CursPos% - 1) / ScreenWidth%)
                ColCurs% = (Lonna% + CursPos% - 1) MOD ScreenWidth%
                IF ColCurs% = 0 THEN ColCurs% = ScreenWidth%: RowCurs% = RowCurs% - 1
                VirtPos% = (Riga% * ScreenWidth% - 1) + Lonna% + FieldLen%
                IF VirtPos% > TotalScreenLength% THEN
                Riga% = Riga% - (INT((VirtPos% - TotalScreenLength% - 1) / ScreenWidth%))

                LOCATE Riga%, Lonna%, 0
                END IF
                LOCATE Riga% + RowCurs%, ColCurs%, 1, CurShape%, 7
                DO
                KeyStroke$ = INKEY$
                LOOP WHILE KeyStroke$ = ""


          If INSTR(AllowCharsMask$, KeyStroke$) > 0 THEN
                Mark% = 0'No Backspace at Posit. 0
                IF CurShape% = standard% THEN
                MID$(ReturnVar$, CursPos%, 1) = KeyStroke$

                ELSE
                ReturnVar$ = LEFT$(ReturnVar$, CursPos% - 1) + KeyStroke$ + MID$(ReturnVar$, CursPos%, FieldLen% - CursPos%)
                END IF

                IF CursPos% < FieldLen% THEN
                CursPos% = CursPos% + 1
                ELSE
                IF MyVal(1) <> 0 AND FieldLen%>1 THEN BEEP
                END IF
                ELSE

                IF INSTR(EndingKeys$, KeyStroke$) > 0 THEN
                ExitKey$ = KeyStroke$
                NotDone% = 0
                ELSE
'  Check for other keys  

                SELECT CASE KeyStroke$
                CASE CursLeft$
                IF CursPos% > 1 THEN
                CursPos% = CursPos% - 1
                END IF
                CASE CursRight$
                IF CursPos% < FieldLen% THEN
                CursPos% = CursPos% + 1
                END IF
                CASE BackS$
                IF CursPos% > 1 THEN
                Mark% = 0'No Backspace at Posit. 0
                ReturnVar$ = LEFT$(ReturnVar$, CursPos% - 2) + RIGHT$(ReturnVar$, FieldLen% - CursPos% + 1) + EmptySpaceChar$
                CursPos% = CursPos% - 1
                ELSE
                Mark% = Mark% + 1:
                SOUND Mark% * 400 + 500, 1
                IF Mark% > 7 THEN
                Mark% = 0
                ReturnVar$ = Default$
                ' IF LEN(ReturnVar$) < FieldLen% THEN
                '    ReturnVar$ = ReturnVar$ + SPACE$(FieldLen% - LEN(ReturnVar$))
                ' END IF
                END IF
                END IF
                CASE Delete$
                ReturnVar$ = LEFT$(ReturnVar$, CursPos% - 1) + RIGHT$(ReturnVar$, FieldLen% - CursPos%) + EmptySpaceChar$
                CASE Insert$, CtrlR$
                IF CurShape% = standard% THEN
                CurShape% = Insert%
                ELSE CurShape% = standard%
                END IF

                CASE Home$, CtrlK$  ' Cursor to field start
                CursPos% = 1

                CASE End$, CtrlN$  ' Cursor to field end
                CursPos% = LEN(ReturnVar$)

                CASE CursUp$  ' Cursor 1 line up
                IF CursPos% > ScreenWidth% THEN CursPos% = CursPos% - ScreenWidth%

                CASE CursDown$ ' Cursor 1 line down
                IF CursPos% + ScreenWidth% < LEN(ReturnVar$) + 1 THEN CursPos% = CursPos% + ScreenWidth%

                CASE CtrlHome$  ' Cancel all chars from Cursor Position -1  up to start
                ReturnVar$ = RIGHT$(ReturnVar$, FieldLen% - CursPos% + 1) + STRING$(CursPos% - 1, EmptySpaceChar$)
                CursPos% = 1

                CASE CtrlEnd$, CtrlE$ ' Ctrl+ E = Cancel all chars from Cursor Posit. up to end
                ReturnVar$ = LEFT$(ReturnVar$, CursPos% - 1) + STRING$(FieldLen% - CursPos% + 1, EmptySpaceChar$)

                CASE PageUp$  ' Cursor to 1st line
                DO WHILE CursPos% > ScreenWidth%
                CursPos% = CursPos% - ScreenWidth%
                LOOP

                CASE PageDown$ ' Cursor to last line
                DO WHILE CursPos% + ScreenWidth% < LEN(ReturnVar$) + 1
                CursPos% = CursPos% + ScreenWidth%
                LOOP

                CASE CtrlCursLeft$, CtrlB$ ' Ctrl+ B = Cursor to the word on the left
                IF CursPos% > 1 THEN
                P% = CursPos% - 1
                DO WHILE (P% > 1) AND INSTR(Separator$, MID$(ReturnVar$, P%, 1)) > 0
                P% = P% - 1
                LOOP

                DO WHILE (P% > 1) AND INSTR(Separator$, MID$(ReturnVar$, P%, 1)) = 0
                P% = P% - 1
                LOOP
                CursPos% = P%
                IF CursPos% < 1 THEN CursPos% = 1
                IF CursPos% > 1 THEN CursPos% = CursPos% + 1
                END IF

                CASE CtrlCursRight$, CtrlF$ ' Ctrl+ F = Cursor to the word on the right
                P% = CursPos%
                DO WHILE (P% < FieldLen%) AND INSTR(Separator$, MID$(ReturnVar$, P%, 1)) = 0
                P% = P% + 1
                LOOP
                DO WHILE (P% < FieldLen%) AND INSTR(Separator$, MID$(ReturnVar$, P%, 1)) > 0
                P% = P% + 1
                LOOP

                CursPos% = P%
                IF CursPos% > FieldLen% THEN CursPos% = FieldLen%

                CASE Esc$, CtrlU$  ' Ctrl+ U = Clear all input field, if Escape is not
                ' activate to exit by XINPUT caller
                ReturnVar$ = STRING$(FieldLen%, EmptySpaceChar$)
                CursPos% = 1

                CASE Tabul$  ' Tabulator in QBasic works in the following way:
                ' Cursor moves to past the NEXT 8th character,
                '   character 1 in string being 1st character.
                ' I.e., to 9th, 17th, 25th, 33rd character and so on.
                ' If INSERT key is not active, up to 8 character will be blanked.
                ' If INSERT key is ACTIVE, up to 8 blanks will be INSERTED.

                P% = (INT((CursPos% - 1 + TabLen%) / TabLen%) * TabLen%) + 1 ' Find Cursor position of next tabulation
                IF P% > FieldLen% + 1 THEN P% = FieldLen% + 1
                IF CursPos% > 1 THEN R1$ = LEFT$(ReturnVar$, CursPos% - 1) ELSE R1$ = ""
                IF CurShape% = standard% THEN
                IF CursPos% < P% THEN R2$ = RIGHT$(ReturnVar$, FieldLen% - P% + 1) ELSE R2$ = ""
                ELSE
                IF CursPos% < FieldLen% THEN R2$ = MID$(ReturnVar$, CursPos%, FieldLen% + 1 - P%) ELSE R2$ = ""
                END IF
                ReturnVar$ = R1$ + STRING$(P% - CursPos%, CHR$(32)) + R2$
                IF P% > FieldLen% THEN P% = FieldLen%
                CursPos% = P%
                
		Case CTRLC$  ' CTRL+ C
		If Len(ReturnVar$)>0 Then MyName$(13)=ReturnVar$
		    
		Case CTRLV$  ' CTRL+ V
		If Len(MyName$(13))=0 Then GoTo SelectEnd
		If Len(MyName$(13)) > Len(ReturnVar$) Then ReturnVar$= left$(MyName$(13),Len(ReturnVar$)) Else ReturnVar$=MyName$(13)		
		CursPos% = 1
		    
		Case ELSE
		SOUND 500, 1
		End SELECT
SelectEnd:	
            
                END IF
	  End IF
          LOOP WHILE NotDone%

'  Exit from here  

                RealStrLen% = LEN(ReturnVar$)
                Xinput$ = ReturnVar$
                END FUNCTION

